import typing

import attr
from applitools.common import FloatingBounds, MatchLevel, Region, logger

from .region import (
    FloatingRegionByRectangle,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)

if typing.TYPE_CHECKING:
    from typing import Text, Optional, List, Union, Tuple
    from applitools.common.utils.custom_types import AnyWebElement, Num

    REGION_VALUES = Union[Region, Text, AnyWebElement, Tuple[Text, Text]]
    FLOATING_VALUES = Union[Region, Text, AnyWebElement, Tuple[Text, Text]]
    GR = typing.TypeVar("GR", bound=GetRegion)
    FR = typing.TypeVar("FR", bound=GetFloatingRegion)

__all__ = ("CheckSettings", "CheckSettingsValues")


@attr.s
class CheckSettingsValues(object):
    """
    Access to values stored in :py:class:`CheckSettings`
    """

    target_region = attr.ib(init=False, default=None)  # type: Optional[Region]
    timeout = attr.ib(init=False, default=-1)  # type: Num  # milliseconds

    ignore_caret = attr.ib(init=False, default=None)  # type: bool
    stitch_content = attr.ib(init=False, default=None)  # type: bool
    match_level = attr.ib(init=False, default=None)  # type: Optional[MatchLevel]
    name = attr.ib(init=False, default=None)  # type: Optional[Text]

    send_dom = attr.ib(init=False, default=None)  # type: bool
    use_dom = attr.ib(init=False, default=None)
    enable_patterns = attr.ib(init=False, default=None)
    ignore_displacement = attr.ib(init=False, default=None)

    ignore_regions = attr.ib(init=False, factory=list)  # type: List[GetRegion]
    layout_regions = attr.ib(init=False, factory=list)  # type: List[GetRegion]
    strict_regions = attr.ib(init=False, factory=list)  # type: List[GetRegion]
    content_regions = attr.ib(init=False, factory=list)  # type: List[GetRegion]
    floating_regions = attr.ib(
        init=False, factory=list
    )  # type: List[GetFloatingRegion]


@attr.s
class CheckSettings(object):
    """
    The Match settings object to use in the various Eyes.Check methods.
    """

    values = attr.ib(init=False)

    def __attrs_post_init__(self):
        # type: () -> None
        self.values = CheckSettingsValues()

    def layout(self, *args):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Shortcut to set the match level to :py:attr:`MatchLevel.LAYOUT`. """
        self.values.match_level = MatchLevel.LAYOUT
        if not args:
            return self
        return self._layout_regions(*args)

    def exact(self):
        # type: ()  -> CheckSettings

        """ Shortcut to set the match level to :py:attr:`MatchLevel.EXACT`. """
        self.values.match_level = MatchLevel.EXACT
        return self

    def strict(self, *args):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Shortcut to set the match level to :py:attr:`MatchLevel.STRICT`. """
        self.values.match_level = MatchLevel.STRICT
        if not args:
            return self
        return self._strict_regions(*args)

    def content(self, *args):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Shortcut to set the match level to :py:attr:`MatchLevel.CONTENT`. """
        self.values.match_level = MatchLevel.CONTENT
        if not args:
            return self
        return self._content_regions(*args)

    def match_level(self, match_level):
        # type: (MatchLevel)  -> CheckSettings
        self.values.match_level = match_level
        return self

    def ignore_caret(self, ignore=True):
        # type: (bool)  -> CheckSettings
        self.values.ignore_caret = ignore
        return self

    def fully(self, fully=True):
        # type: (bool)  -> CheckSettings
        self.values.stitch_content = fully
        return self

    def with_name(self, name):
        # type: (Text)  -> CheckSettings
        self.values.name = name
        return self

    def stitch_content(self, stitch_content=True):
        # type: (bool)  -> CheckSettings
        self.values.stitch_content = stitch_content
        return self

    def timeout(self, timeout):
        # type: (int)  -> CheckSettings
        self.values.timeout = timeout
        return self

    def ignore_regions(self, *regions):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Adds one or more ignore regions. """
        self.values.ignore_regions = self.__regions(
            regions, method_name="ignore_regions"
        )
        return self

    ignore = ignore_regions

    def _layout_regions(self, *regions):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Adds one or more layout regions. """
        self.values.layout_regions = self.__regions(
            regions, method_name="layout_regions"
        )
        return self

    def _strict_regions(self, *regions):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Adds one or more strict regions. """
        self.values.strict_regions = self.__regions(
            regions, method_name="strict_regions"
        )
        return self

    def _content_regions(self, *regions):
        # type: (*REGION_VALUES)  -> CheckSettings
        """ Adds one or more content regions. """
        self.values.content_regions = self.__regions(
            regions, method_name="content_regions"
        )
        return self

    def floating_region(
        self,
        arg1,  # type: Union[REGION_VALUES, int]
        arg2,  # type: Union[REGION_VALUES, int]
        arg3=None,  # type: Optional[int]
        arg4=None,  # type: Optional[int]
        arg5=None,  # type: Optional[int]
    ):
        # type: (...) -> CheckSettings
        """
        Adds a floating region. Region and max_offset or [max_up_offset, max_down_offset, "
                "max_left_offset, max_right_offset] are required parameters.

        :param arg1: max_offset | Region
        :param arg2: Region     | max_up_offset
        :param arg3: None       | max_down_offset
        :param arg4: None       | max_left_offset
        :param arg5: None       | max_right_offset
        """
        if isinstance(arg1, int) and isinstance(arg2, Region):
            max_offset = arg1  # type: int
            region = arg2  # type: Region
            bounds = FloatingBounds(
                max_up_offset=max_offset,
                max_down_offset=max_offset,
                max_left_offset=max_offset,
                max_right_offset=max_offset,
            )
        elif (
            isinstance(arg2, int)
            and isinstance(arg3, int)
            and isinstance(arg4, int)
            and isinstance(arg5, int)
        ):
            region = arg1
            bounds = FloatingBounds(
                max_up_offset=arg2,
                max_down_offset=arg3,
                max_left_offset=arg4,
                max_right_offset=arg5,
            )
        else:
            raise TypeError("Unsupported parameters")
        logger.info("Adding Region {} with FloatingBounds {}".format(region, bounds))
        region_or_container = self._floating_provider_from(region, bounds)
        self.values.floating_regions.append(region_or_container)
        return self

    floating = floating_region

    def send_dom(self, send=True):
        # type: (bool) -> CheckSettings
        """
         Defines whether to send the document DOM or not.
        """
        self.values.send_dom = send
        return self

    def use_dom(self, use=True):
        # type: (bool) -> CheckSettings
        """
         Defines useDom for enabling the match algorithm to use dom.
        """
        self.values.use_dom = use
        return self

    def enable_patterns(self, enable=True):
        # type: (bool) -> CheckSettings
        self.values.enable_patterns = enable
        return self

    def ignore_displacement(self, should_ignore=True):
        # type: (bool) -> CheckSettings
        self.values.ignore_displacement = should_ignore
        return self

    def __regions(self, regions, method_name):
        if not regions:
            raise TypeError(
                "{name} method called without arguments!".format(name=method_name)
            )

        regions_list = getattr(self.values, method_name)
        for region in regions:
            regions_list.append(self._region_provider_from(region, method_name))
        return regions_list

    def _region_provider_from(self, region, method_name):
        logger.debug("calling _{}".format(method_name))
        if isinstance(region, Region):
            logger.debug("{name}: IgnoreRegionByElement".format(name=method_name))
            return RegionByRectangle(region)
        raise TypeError("Unknown region type.")

    def _floating_provider_from(self, region, bounds):
        if isinstance(region, Region):
            logger.debug("floating: FloatingRegionByRectangle")
            return FloatingRegionByRectangle(Region.from_(region), bounds)
        raise TypeError("Unknown region type.")



from __future__ import absolute_import

import json
import math
import time
import typing
from struct import pack

import attr
import requests
from requests import Response
from requests.packages import urllib3  # noqa

from applitools.common import RenderStatus, RunningSession, logger
from applitools.common.errors import EyesError
from applitools.common.match import MatchResult
from applitools.common.match_window_data import MatchWindowData
from applitools.common.metadata import SessionStartInfo
from applitools.common.test_results import TestResults
from applitools.common.utils import (
    argument_guard,
    datetime_utils,
    gzip_compress,
    image_utils,
    json_utils,
    urljoin,
)
from applitools.common.utils.general_utils import retry
from applitools.common.visual_grid import (
    RenderingInfo,
    RenderRequest,
    RenderStatusResults,
    RunningRender,
    VGResource,
)

if typing.TYPE_CHECKING:
    from typing import Text, List, Any, Optional, Dict, Tuple
    from applitools.common.utils.custom_types import Num

# Prints out all data sent/received through 'requests'
# import httplib
# httplib.HTTPConnection.debuglevel = 1

# Remove Unverified SSL warnings propagated by requests' internal urllib3 module
if hasattr(urllib3, "disable_warnings") and callable(urllib3.disable_warnings):
    urllib3.disable_warnings()

__all__ = ("ServerConnector",)


@attr.s
class _RequestCommunicator(object):
    LONG_REQUEST_DELAY_SEC = 2  # type: int
    MAX_LONG_REQUEST_DELAY_SEC = 10  # type: int
    LONG_REQUEST_DELAY_MULTIPLICATIVE_INCREASE_FACTOR = 1.5  # type: float

    headers = attr.ib()  # type: Dict
    timeout_sec = attr.ib(default=None)  # type: int
    api_key = attr.ib(default=None)  # type: Text
    server_url = attr.ib(default=None)  # type: Text

    def request(self, method, url_resource, use_api_key=True, **kwargs):
        if url_resource is not None:
            # makes URL relative
            url_resource = url_resource.lstrip("/")
        url_resource = urljoin(self.server_url, url_resource)
        params = {}
        if use_api_key:
            params["apiKey"] = self.api_key
        params.update(kwargs.get("params", {}))
        headers = kwargs.get("headers", self.headers).copy()
        timeout = kwargs.get("timeout", self.timeout_sec)
        response = method(
            url_resource,
            data=kwargs.get("data", None),
            verify=False,
            params=params,
            headers=headers,
            timeout=timeout,
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.exception(e)
        return response

    def long_request(self, method, url_resource, **kwargs):
        headers = kwargs.get("headers", self.headers).copy()
        headers["Eyes-Expect"] = "202+location"
        headers["Eyes-Date"] = datetime_utils.current_time_in_rfc1123()
        kwargs["headers"] = headers
        response = self.request(method, url_resource, **kwargs)
        return self._long_request_check_status(response)

    def _long_request_check_status(self, response):
        if response.status_code == requests.codes.ok:
            # request ends successful
            return response
        elif response.status_code == requests.codes.accepted:
            # long request here; calling received url to know that request was processed
            url = response.headers["Location"]
            response = self._long_request_loop(url)
            return self._long_request_check_status(response)
        elif response.status_code == requests.codes.created:
            # delete url that was used before
            url = response.headers["Location"]
            return self.request(
                requests.delete,
                url,
                headers={"Eyes-Date": datetime_utils.current_time_in_rfc1123()},
            )
        elif response.status_code == requests.codes.gone:
            raise EyesError("The server task has gone.")
        else:
            raise EyesError("Unknown error during long request: {}".format(response))

    def _long_request_loop(self, url, delay=LONG_REQUEST_DELAY_SEC):
        delay = min(
            self.MAX_LONG_REQUEST_DELAY_SEC,
            math.floor(delay * self.LONG_REQUEST_DELAY_MULTIPLICATIVE_INCREASE_FACTOR),
        )
        logger.debug("Still running... Retrying in {}s".format(delay))

        time.sleep(delay)
        response = self.request(
            requests.get,
            url,
            headers={"Eyes-Date": datetime_utils.current_time_in_rfc1123()},
        )
        if response.status_code != requests.codes.ok:
            return response
        return self._long_request_loop(url, delay)


def prepare_match_data(match_data):
    # type: (MatchWindowData) -> bytes
    screenshot64 = match_data.app_output.screenshot64
    if screenshot64:
        match_data.app_output.screenshot64 = None
        image = image_utils.image_from_base64(screenshot64)
        screenshot_bytes = image_utils.get_bytes(image)  # type: bytes
    else:
        screenshot_bytes = b""
    match_data_json = json_utils.to_json(match_data)
    logger.debug("MatchWindowData {}".format(match_data_json))
    match_data_json_bytes = match_data_json.encode("utf-8")  # type: bytes
    match_data_size_bytes = pack(">L", len(match_data_json_bytes))  # type: bytes
    body = match_data_size_bytes + match_data_json_bytes + screenshot_bytes
    return body


class ServerConnector(object):
    """
    Provides an API for communication with the Applitools server.
    """

    DEFAULT_HEADERS = {"Accept": "application/json", "Content-Type": "application/json"}

    API_SESSIONS = "api/sessions"
    API_SESSIONS_RUNNING = API_SESSIONS + "/running/"
    RUNNING_DATA_PATH = API_SESSIONS + "/running/data"

    # Rendering Grid
    RENDER_INFO_PATH = API_SESSIONS + "/renderinfo"
    RESOURCES_SHA_256 = "/resources/sha256/"
    RENDER_STATUS = "/render-status"
    RENDER = "/render"

    _is_session_started = False

    def __init__(self):
        # type: () -> None
        """
        Ctor.

        :param server_url: The url of the Applitools server.
        """
        self._render_info = None  # type: Optional[RenderingInfo]
        self._com = _RequestCommunicator(headers=ServerConnector.DEFAULT_HEADERS)

    def update_config(self, conf):
        if conf.api_key is None:
            raise EyesError(
                "API key not set! Log in to https://applitools.com to obtain your"
                " API Key and use 'api_key' to set it."
            )
        self._com.server_url = conf.server_url
        self._com.api_key = conf.api_key
        self._com.timeout_sec = conf.timeout / 1000.0

    @property
    def server_url(self):
        return self._com.server_url

    @server_url.setter
    def server_url(self, value):
        self._com.server_url = value

    @property
    def api_key(self):
        return self._com.api_key

    @api_key.setter
    def api_key(self, value):
        self._com.api_key = value

    @property
    def timeout(self):
        return self._com.timeout_sec * 1000  # ms

    @timeout.setter
    def timeout(self, value):
        self._com.timeout_sec = value / 1000.0

    @property
    def is_session_started(self):
        return self._is_session_started

    # TODO: Add Proxy
    def start_session(self, session_start_info):
        # type: (SessionStartInfo) -> RunningSession
        """
        Starts a new running session in the agent. Based on the given parameters,
        this running session will either be linked to an existing session, or to
        a completely new session.

        :param session_start_info: The start params for the session.
        :return: Represents the current running session.
        """
        logger.debug("start_session called.")
        data = json_utils.to_json(session_start_info)
        response = self._com.request(
            requests.post, url_resource=self.API_SESSIONS_RUNNING, data=data
        )
        running_session = json_utils.attr_from_response(response, RunningSession)
        running_session.is_new_session = response.status_code == requests.codes.created
        self._is_session_started = True
        return running_session

    def stop_session(self, running_session, is_aborted, save):
        # type: (RunningSession, bool, bool) -> TestResults
        """
        Stops a running session in the Eyes server.

        :param running_session: The session to stop.
        :param is_aborted: Whether the server should mark this session as aborted.
        :param save: Whether the session should be automatically saved if it is not aborted.
        :return: Test results of the stopped session.
        """
        logger.debug("stop_session called.")

        if not self.is_session_started:
            raise EyesError("Session not started")

        params = {"aborted": is_aborted, "updateBaseline": save}
        response = self._com.long_request(
            requests.delete,
            url_resource=urljoin(self.API_SESSIONS_RUNNING, running_session.id),
            params=params,
            headers=ServerConnector.DEFAULT_HEADERS,
        )

        test_results = json_utils.attr_from_response(response, TestResults)
        logger.debug("stop_session(): parsed response: {}".format(test_results))

        # mark that session isn't started
        self._is_session_started = False
        return test_results

    def match_window(self, running_session, match_data):
        # type: (RunningSession, MatchWindowData) -> MatchResult
        """
        Matches the current window to the immediate expected window in the Eyes server.
        Notice that a window might be matched later at the end of the test, even if it
        was not immediately matched in this call.

        :param running_session: The current session that is running.
        :param match_data: The data for the requests.post.
        :return: The parsed response.
        """
        logger.debug("match_window called. {}".format(running_session))

        # logger.debug("Data length: %d, data: %s" % (len(data), repr(data)))
        if not self.is_session_started:
            raise EyesError("Session not started")

        data = prepare_match_data(match_data)
        # Using the default headers, but modifying the "content type" to binary
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/octet-stream"
        # TODO: allow to send images as base64
        response = self._com.long_request(
            requests.post,
            url_resource=urljoin(self.API_SESSIONS_RUNNING, running_session.id),
            data=data,
            headers=headers,
        )
        match_result = json_utils.attr_from_response(response, MatchResult)
        return match_result

    def post_dom_snapshot(self, dom_json):
        # type: (Text) -> Optional[Text]
        """
        Upload the DOM of the tested page.
        Return an URL of uploaded resource which should be posted to :py:   `AppOutput`.
        """
        logger.debug("post_dom_snapshot called.")

        if not self.is_session_started:
            raise EyesError("Session not started")

        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/octet-stream"
        dom_bytes = gzip_compress(dom_json.encode("utf-8"))

        response = self._com.request(
            requests.post,
            url_resource=urljoin(self.API_SESSIONS_RUNNING, "data"),
            data=dom_bytes,
            headers=headers,
        )
        dom_url = None
        if response.ok:
            dom_url = response.headers["Location"]
        return dom_url

    def render_info(self):
        # type: () -> Optional[RenderingInfo]
        logger.debug("render_info() called.")
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/json"
        response = self._com.request(
            requests.get, self.RENDER_INFO_PATH, headers=headers
        )
        if not response.ok:
            raise EyesError(
                "Cannot get render info: \n Status: {}, Content: {}".format(
                    response.status_code, response.content
                )
            )
        self._render_info = json_utils.attr_from_response(response, RenderingInfo)
        return self._render_info

    def render(self, *render_requests):
        # type: (*RenderRequest) -> List[RunningRender]
        logger.debug("render called with {}".format(render_requests))
        if self._render_info is None:
            raise EyesError("render_info must be fetched first")

        url = urljoin(self._render_info.service_url, self.RENDER)

        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/json"
        headers["X-Auth-Token"] = self._render_info.access_token

        data = json_utils.to_json(render_requests)
        response = self._com.request(
            requests.post, url, use_api_key=False, headers=headers, data=data
        )
        if response.ok or response.status_code == requests.codes.not_found:
            return json_utils.attr_from_response(response, RunningRender)
        raise EyesError(
            "ServerConnector.render - unexpected status ({})\n\tcontent{}".format(
                response.status_code, response.content
            )
        )

    def render_put_resource(self, running_render, resource):
        # type: (RunningRender, VGResource) -> Text
        argument_guard.not_none(running_render)
        argument_guard.not_none(resource)
        if self._render_info is None:
            raise EyesError("render_info must be fetched first")

        content = resource.content
        argument_guard.not_none(content)
        logger.debug(
            "resource hash: {} url: {} render id: {}"
            "".format(resource.hash, resource.url, running_render.render_id)
        )
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = resource.content_type
        headers["X-Auth-Token"] = self._render_info.access_token

        url = urljoin(
            self._render_info.service_url, self.RESOURCES_SHA_256 + resource.hash
        )
        response = self._com.request(
            requests.put,
            url,
            use_api_key=False,
            headers=headers,
            data=content,
            params={"render-id": running_render.render_id},
        )
        logger.debug("ServerConnector.put_resource - request succeeded")
        if not response.ok:
            raise EyesError(
                "Error putting resource: {}, {}".format(
                    response.status_code, response.content
                )
            )
        return resource.hash

    @retry()
    def download_resource(self, url):
        # type: (Text) -> Response
        # TODO: Fixme retry mech
        logger.debug("Fetching {}...".format(url))
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Accept-Encoding"] = "identity"

        response = requests.get(
            url, headers=headers, timeout=self._com.timeout_sec, verify=False
        )
        if response.status_code == requests.codes.not_acceptable:
            response = requests.get(url, timeout=self._com.timeout_sec, verify=False)
        response.raise_for_status()
        return response

    def render_status_by_id(self, render_id):
        # type: (Text) -> List[RenderStatusResults]
        argument_guard.not_none(render_id)
        if self._render_info is None:
            raise EyesError("render_info must be fetched first")

        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/json"
        headers["X-Auth-Token"] = self._render_info.access_token
        url = urljoin(self._render_info.service_url, self.RENDER_STATUS)
        response = self._com.request(
            requests.post,
            url,
            use_api_key=False,
            headers=headers,
            data=json.dumps([render_id]),
        )
        if not response.ok:
            raise EyesError(
                "Error getting server status, {} {}".format(
                    response.status_code, response.content
                )
            )
        return json_utils.attr_from_response(response, RenderStatusResults)



import abc
import typing

import attr

from applitools.common.geometry import Point
from applitools.common.utils import ABC, iteritems

if typing.TYPE_CHECKING:
    from typing import List, Optional
    from applitools.common import RectangleSize


class PositionMomento(object):
    """
    A base class for position related memento instances. This is intentionally
    not an interface, since the mementos might vary in their interfaces.
    """

    def __init__(self, position, **kwargs):
        self.position = position
        for name, val in iteritems(kwargs):
            setattr(self, name, val)

    def __str__(self):
        return "PositionMomento(position={}, ...)".format(self.position)


@attr.s
class PositionProvider(ABC):
    _states = attr.ib(init=False, factory=list)  # type: List[PositionMomento]
    _last_set_position = attr.ib(init=False, default=None)  # type: Point

    @abc.abstractmethod
    def get_current_position(self):
        # type: () -> Optional[Point]
        """
        :return: The current position, or `None` if position is not available.
        """

    @abc.abstractmethod
    def set_position(self, location):
        # type: (Point) -> Point
        """
        Go to the specified location.

        :param location: The position to set
        """

    @abc.abstractmethod
    def get_entire_size(self):
        # type: () -> RectangleSize
        """
        :return: The entire size of the container which the position is relative to.
        """

    def push_state(self):
        """
        Adds the current position to the states list.
        """
        self._states.append(PositionMomento(self.get_current_position()))

    def pop_state(self):
        """
        Sets the position to be the last position added to the states list.
        """
        state = self._states.pop()
        self.set_position(state.position)
        self._last_set_position = state.position

    @property
    def states(self):
        # type: () -> List[Point]
        return self._states

    def __enter__(self):
        self.push_state()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop_state()
        return self


class InvalidPositionProvider(PositionProvider):
    """
    An implementation of :py:class:`PositionProvider` which throws an exception
    for every method. Can be used as a placeholder until an actual
    implementation is set.
    """

    def get_current_position(self):
        raise NotImplementedError("This class does not implement methods!")

    def set_position(self, location):
        # type: (Point) -> None
        raise NotImplementedError("This class does not implement methods!")

    def get_entire_size(self):
        raise NotImplementedError("This class does not implement methods!")
